from __future__ import annotations
#  Copyright 2020, 2021 Evandro Chagas Ribeiro da Rosa <evandro.crr@posgrad.ufsc.br>
#  Copyright 2020, 2021 Rafael de Santiago <r.santiago@ufsc.br>
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from ..base import label, branch, jump, future


__all__ = ['_ket_is_future', '_ket_if', '_ket_if_else', '_ket_else', '_ket_next',
           '_ket_while', '_ket_while_else', '_ket_while_body', '_ket_loop', '_ket_goto']


def _ket_is_future(obj) -> bool:
    return isinstance(obj, future)


def _ket_if(test: future) -> label:
    if_then = label()
    if_end = label()
    branch(test, if_then, if_end)
    if_then.begin()
    return if_end


def _ket_if_else(test: future) -> tuple[label]:
    if_then = label()
    if_else = label()
    if_end = label()
    branch(test, if_then, if_else)
    if_then.begin()
    return if_else, if_end


def _ket_else(else_: label, end: label):
    jump(end)
    else_.begin()


def _ket_next(end: label):
    jump(end)
    end.begin()


def _ket_while() -> tuple[label]:
    while_test = label()
    while_loop = label()
    while_end = label()
    jump(while_test)
    while_test.begin()
    return while_test, while_loop, while_end


def _ket_while_else() -> tuple[label]:
    while_test = label()
    while_loop = label()
    while_else = label()
    while_end = label()
    jump(while_test)
    while_test.begin()
    return while_test, while_loop, while_else, while_end


def _ket_while_body(test: future, body: label, end: label):
    branch(test, body, end)
    body.begin()


def _ket_loop(test: label, end: label):
    jump(test)
    end.begin()


def _ket_goto(where: label):
    dead_code = label()
    jump(where)
    dead_code.begin()



from __future__ import annotations
#  Copyright 2020, 2021 Evandro Chagas Ribeiro da Rosa <evandro.crr@posgrad.ufsc.br>
#  Copyright 2020, 2021 Rafael de Santiago <r.santiago@ufsc.br>
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from ctypes import *
from .wrapper import load_lib, os_lib_name
import weakref

EQ = 0
NEQ = 1
GT = 2
GEQ = 3
LT = 4
LEQ = 5
ADD = 6
SUB = 7
MUL = 8
MOD = 9
DIV = 10
SLL = 11
SRL = 12
AND = 13
OR = 14
XOR = 15

PAULI_X = 0
PAULI_Y = 1
PAULI_Z = 2
HADAMARD = 3
PHASE = 4
RX = 5
RY = 6
RZ = 7

SUCCESS = 0
CONTROL_TWICE = 1
DATA_NOT_AVAILABLE = 2
DEALLOCATED_QUBIT = 3
FAIL_TO_PARSE_RESULT = 4
NO_ADJ = 5
NO_CTRL = 6
NON_GATE_INSTRUCTION = 7
NOT_BIN = 8
NOT_JSON = 9
NOT_UNITARY = 10
PLUGIN_ON_CTRL = 11
TARGET_ON_CONTROL = 12
TERMINATED_BLOCK = 13
UNDEFINED_CLASSICAL_OP = 14
UNDEFINED_DATA_TYPE = 15
UNDEFINED_GATE = 16
UNEXPECTED_RESULT_DATA = 17
UNMATCHED_PID = 18
UNDEFINED_ERROR = 19

JSON = 0
BIN = 1

API_argtypes = {
    # 'ket_type_method': ([input_list], [output_list]),
    'ket_process_new': ([c_size_t], [c_void_p]),
    'ket_process_delete': ([c_void_p], []),
    'ket_process_allocate_qubit': ([c_void_p, c_bool], [c_void_p]),
    'ket_process_free_qubit': ([c_void_p, c_void_p, c_bool], []),
    'ket_process_apply_gate': ([c_void_p, c_int32, c_double, c_void_p], []),
    'ket_process_apply_plugin': ([c_void_p, POINTER(c_char), POINTER(c_char), POINTER(c_void_p), c_size_t], []),
    'ket_process_measure': ([c_void_p, POINTER(c_void_p), c_size_t], [c_void_p]),
    'ket_process_ctrl_push': ([c_void_p, POINTER(c_void_p), c_size_t], []),
    'ket_process_ctrl_pop': ([c_void_p], []),
    'ket_process_adj_begin': ([c_void_p], []),
    'ket_process_adj_end': ([c_void_p], []),
    'ket_process_get_label': ([c_void_p], [c_void_p]),
    'ket_process_open_block': ([c_void_p, c_void_p], []),
    'ket_process_jump': ([c_void_p, c_void_p], []),
    'ket_process_branch': ([c_void_p, c_void_p, c_void_p, c_void_p], []),
    'ket_process_dump': ([c_void_p, POINTER(c_void_p), c_size_t], [c_void_p]),
    'ket_process_add_int_op': ([c_void_p, c_int32, c_void_p, c_void_p], [c_void_p]),
    'ket_process_int_new': ([c_void_p, c_int64], [c_void_p]),
    'ket_process_int_set': ([c_void_p, c_void_p, c_void_p], []),
    'ket_process_prepare_for_execution': ([c_void_p], []),
    'ket_process_exec_time': ([c_void_p], [c_double]),
    'ket_process_set_timeout': ([c_void_p, c_uint64], []),
    'ket_process_serialize_metrics': ([c_void_p, c_int32], []),
    'ket_process_serialize_quantum_code': ([c_void_p, c_int32], []),
    'ket_process_get_serialized_metrics': ([c_void_p], [POINTER(c_uint8), c_size_t, c_int32]),
    'ket_process_get_serialized_quantum_code': ([c_void_p], [POINTER(c_uint8), c_size_t, c_int32]),
    'ket_process_set_serialized_result': ([c_void_p, POINTER(c_uint8), c_size_t, c_int32], []),
    'ket_qubit_delete': ([c_void_p], []),
    'ket_qubit_index': ([c_void_p], [c_size_t]),
    'ket_qubit_pid': ([c_void_p], [c_size_t]),
    'ket_qubit_allocated': ([c_void_p], [c_bool]),
    'ket_qubit_measured': ([c_void_p], [c_bool]),
    'ket_dump_delete': ([c_void_p], []),
    'ket_dump_states_size': ([c_void_p], [c_size_t]),
    'ket_dump_state': ([c_void_p, c_size_t], [POINTER(c_uint64), c_size_t]),
    'ket_dump_amplitudes_real': ([c_void_p], [POINTER(c_double), c_size_t]),
    'ket_dump_amplitudes_imag': ([c_void_p], [POINTER(c_double), c_size_t]),
    'ket_dump_available': ([c_void_p], [c_bool]),
    'ket_future_delete': ([c_void_p], []),
    'ket_future_value': ([c_void_p], [c_int64]),
    'ket_future_index': ([c_void_p], [c_size_t]),
    'ket_future_pid': ([c_void_p], [c_size_t]),
    'ket_future_available': ([c_void_p], [c_bool]),
    'ket_label_delete': ([c_void_p], []),
    'ket_label_index': ([c_void_p], [c_size_t]),
    'ket_label_pid': ([c_void_p], [c_size_t]),
}


def libket_path():
    from os import environ, name
    from os.path import dirname

    if "LIBKET_PATH" in environ:
        libket_path = environ["LIBKET_PATH"]
    else:
        libket_path = f'{dirname(__file__)}/libs/{os_lib_name("ket")}'

    return libket_path


API = load_lib('Libket', libket_path(), API_argtypes, 'ket_error_message')


class process:
    def __init__(self, pid: int):
        self.pid = pid
        self._as_parameter_ = API['ket_process_new'](pid)
        self._finalizer = weakref.finalize(
            self, API['ket_process_delete'], self._as_parameter_)

    def __getattr__(self, name: str):
        return lambda *args: API['ket_process_'+name](self, *args)

    def __repr__(self) -> str:
        return f"<Libket 'process' ({self.pid})>"


class libket_qubit:
    def __init__(self, addr: c_void_p):
        self._as_parameter_ = addr
        self._finalizer = weakref.finalize(
            self, API['ket_qubit_delete'], self._as_parameter_)

    def __getattr__(self, name: str):
        return lambda *args: API['ket_qubit_'+name](self, *args)

    def __repr__(self) -> str:
        return f"<Libket 'qubit' {self.pid().value, self.index().value}>"


class libket_dump:
    def __init__(self, addr: c_void_p):
        self._as_parameter_ = addr
        self._finalizer = weakref.finalize(
            self, API['ket_dump_delete'], self._as_parameter_)

    def __getattr__(self, name: str):
        return lambda *args: API['ket_dump_'+name](self, *args)

    def __repr__(self) -> str:
        return f"<Libket 'dump' {self.pid().value, self.index().value}>"


class libket_future:
    def __init__(self, addr: c_void_p):
        self._as_parameter_ = addr
        self._finalizer = weakref.finalize(
            self, API['ket_future_delete'], self._as_parameter_)

    def __getattr__(self, name: str):
        return lambda *args: API['ket_future_'+name](self, *args)

    def __repr__(self) -> str:
        return f"<Libket 'future' {self.pid().value, self.index().value}>"


class libket_label:
    def __init__(self, addr: c_void_p):
        self._as_parameter_ = addr
        self._finalizer = weakref.finalize(
            self, API['ket_label_delete'], self._as_parameter_)

    def __getattr__(self, name: str):
        return lambda *args: API['ket_label_'+name](self, *args)

    def __repr__(self) -> str:
        return f"<Libket 'label' {self.pid().value, self.index().value}>"



from __future__ import annotations
#  Copyright 2020, 2021 Evandro Chagas Ribeiro da Rosa <evandro.crr@posgrad.ufsc.br>
#  Copyright 2020, 2021 Rafael de Santiago <r.santiago@ufsc.br>
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from functools import reduce
from operator import add
from ..base import base_measure, quant, future, qc_int, base_X
from ..preprocessor import _ket_if, _ket_next  # pylint: disable=no-name-in-module
from .ctrl import *
from .adj import *

__all__ = ['inverse', 'control', 'ctrl',
           'adj', 'around', 'measure', 'qc_int']


def measure(qubits: quant | list[quant], free: bool = False) -> future | list[future]:
    """Quantum measurement

    Measure the qubits of a :class:`~ket.libket.quant` and return a
    :class:`~ket.libket.future` or [:class:`~ket.libket.future`].


    When measuring more than 64 qubits, Ket split the measure every
    63 qubits.

    Args:
        q: Qubits to measure.
        free: If ``True``, free the qubits after the measurement.
    """
    qubits = reduce(add, qubits) if len(qubits) else qubits

    ret = base_measure(qubits)
    if free:
        for i in qubits:
            end = _ket_if(base_measure(i))
            base_X(i)
            _ket_next(end)
        qubits.free()
    return ret



#  Copyright 2020, 2021 Evandro Chagas Ribeiro da Rosa <evandro.crr@posgrad.ufsc.br>
#  Copyright 2020, 2021 Rafael de Santiago <r.santiago@ufsc.br>
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from ctypes import *
import os


def os_lib_name(lib):
    if os.name == 'nt':
        return f'{lib}.dll'
    if os.uname().sysname == 'Linux':
        return f'lib{lib}.so'
    if os.uname().sysname == 'Darwin':
        return f'lib{lib}.dylib'
    raise OSError('unsupported operational system')


def from_u8_to_str(data, size):
    return bytearray(data[:size.value]).decode()


def from_list_to_c_vector(data):
    return (c_void_p*len(data))(*(d._as_parameter_ for d in data)), len(data)


class clib_error(Exception):
    def __init__(self, message, error_code):
        self.error_code = error_code
        self.message = message
        super().__init__(self.message)


class APIWrapper:
    def __init__(self, lib_name, c_call, output, error_message):
        self.lib_name = lib_name
        self.c_call = c_call
        self.output = output
        self.error_message = error_message

    def __call__(self, *args):
        out = [c_type() for c_type in self.output]
        error_code = self.c_call(*args, *out)
        if error_code != 0:
            size = c_size_t()
            error_msg = self.error_message(error_code, size)
            raise clib_error(
                self.lib_name+': '+from_u8_to_str(error_msg, size), error_code)
        elif len(out) == 1:
            return out[0]
        elif len(out) != 0:
            return out


def load_lib(lib_name, lib_path, API_argtypes, error_message):

    lib = cdll.LoadLibrary(lib_path)
    error_message = lib.__getattr__(error_message)
    error_message.argtypes = [c_int32, POINTER(c_size_t)]
    error_message.restype = POINTER(c_uint8)

    API = {}
    for name in API_argtypes:
        c_call = lib.__getattr__(name)
        c_call.argtypes = [
            *API_argtypes[name][0],
            *[POINTER(t) for t in API_argtypes[name][1]]
        ]

        API[name] = APIWrapper(
            lib_name,
            c_call,
            API_argtypes[name][1],
            error_message
        )

    return API



#  Copyright 2020, 2021 Evandro Chagas Ribeiro da Rosa <evandro.crr@posgrad.ufsc.br>
#  Copyright 2020, 2021 Rafael de Santiago <r.santiago@ufsc.br>
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from math import pi
from ..base import base_H, base_RX, base_RZ, base_X, base_Z
from ..standard import ctrl, around, base_flipc

# pylint: disable=missing-function-docstring, invalid-name


def base_cnot(ctrl_qubits, trg_qubits):
    for ctrl_qubit, trg_qubit in zip(ctrl_qubits, trg_qubits):
        ctrl(ctrl_qubit, base_X, trg_qubit)


def base_swap(qubits_a, qubits_b):
    base_cnot(qubits_a, qubits_b)
    base_cnot(qubits_b, qubits_a)
    base_cnot(qubits_a, qubits_b)


def base_RXX(theta, qubits_a, qubits_b):
    for qubit_a, qubit_b in zip(qubits_a, qubits_b):
        with around([lambda a, b: base_H(a+b), base_cnot], qubit_a, qubit_b):
            base_RZ(theta, qubit_b)


def base_RZZ(theta, qubits_a, qubits_b):
    for qubit_a, qubit_b in zip(qubits_a, qubits_b):
        with around(base_cnot, qubit_a, qubit_b):
            base_RZ(theta, qubit_b)


def base_RYY(theta, qubits_a, qubits_b):
    for qubit_a, qubit_b in zip(qubits_a, qubits_b):
        with around([lambda a, b: base_RX(pi/2, a+b), base_cnot], qubit_a, qubit_b):
            base_RZ(theta, qubit_b)


def base_phase_on(state, qubits):
    with around(lambda q: base_flipc(state, q), qubits):
        ctrl(qubits[1:], base_Z, qubits[0])



#  Copyright 2020, 2021 Evandro Chagas Ribeiro da Rosa <evandro.crr@posgrad.ufsc.br>
#  Copyright 2020, 2021 Rafael de Santiago <r.santiago@ufsc.br>
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from .gates import *
from .import_ket import *
from .base import *
from .standard import *
from .process import *
from .gates import __all__ as all_gate
from .import_ket import __all__ as all_import
from .base import __all__ as all_base
from .standard import __all__ as all_standard
from .process import __all__ as all_process


__all__ = all_gate+all_import+all_base+all_standard+all_process

from .import_ket import code_ket

from .base import set_quantum_execution_target, QUANTUM_EXECUTION_TARGET
if QUANTUM_EXECUTION_TARGET is None:
    from .clib.kbw import run_and_set_result
    set_quantum_execution_target(run_and_set_result)



from __future__ import annotations
#  Copyright 2020, 2021 Evandro Chagas Ribeiro da Rosa <evandro.crr@posgrad.ufsc.br>
#  Copyright 2020, 2021 Rafael de Santiago <r.santiago@ufsc.br>
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Callable, Any
from functools import reduce
from operator import add
from ..base import ctrl_push, ctrl_pop, quant, base_X


def base_flipc(state, qubits):  # pylint: disable=missing-function-docstring
    length = len(qubits)
    if hasattr(state, '__iter__'):
        if len(state) != length:
            raise ValueError(
                f"'to' received a list of length {len(state)} to use on {length} qubits")
    else:
        if length < state.bit_length():
            raise ValueError(
                f"To flip with state={state} you need at least {state.bit_length()} qubits")
        state = [int(i) for i in f"{{:0{length}b}}".format(state)]
    for i, qubit in zip(state, qubits):
        if i == 0:
            base_X(qubit)


class control:  # pylint: disable=invalid-name
    r"""Open a controlled-scope

    Inside a ``with control`` scope, the qubits of ``ctr`` control every
    quantum operation, only execution if all qubits are in the estate
    :math:`\left|1\right>`.

    Optionally, you can change the state when applying the quantum operation
    with the Keyword Argument ``on_state``.  If ``on_state`` is an ``int``, the
    number of qubits needs to be equal to or greater than the bit length of
    ``on_state``.  Else if ``on_sate`` is a ``List[int]``, its length needs to
    be equal to the number of qubits.

    For example, execute when the qubits are in the state
    :math:`\left|\left[\dots0\right]11\right>`, use ``with control(ctr, on_state=3):``
    or ``with control(ctr, on_state=[0, 1, 1]):`` if ``ctr`` has exactly 3 qubits.

    :Usage:

    .. code-block:: ket

        with control(*ctr[, on_state]):
            ...

    :Example:

    .. code-block:: ket

        c = quant(2)
        a, b = quant(2)

        # CNOT c[0] a
        with control(c[0]):
            X(a)

        # Toffoli c[0] c[1] a
        with control(c):
            X(a)

        # CSWAP c[0] a b
        with control(c[0]):
            swap(a, b)

    Args:
        ctr: Control qubits.
        on_state: Change the control state.
    """

    def __init__(self, *ctr: quant, on_state: int | list[int] | None = None):
        self.ctr = reduce(add, ctr)
        self.on_state = on_state

    def __enter__(self):
        if self.on_state is not None:
            base_flipc(self.on_state, self.ctr)
        ctrl_push(self.ctr)

    def __exit__(self, type, value, tb):  # pylint: disable=redefined-builtin, invalid-name
        ctrl_pop()
        if self.on_state is not None:
            base_flipc(self.on_state, self.ctr)


def _ctrl(ctrl_qubits: quant | list[quant],
          func: Callable | list[Callable],
          *args,
          on_state: int | list[int] | None = None,
          **kwargs) -> Any:
    """Call Callable with control-qubits"""

    ctrl_qubits = reduce(add, ctrl_qubits)
    if on_state is not None:
        base_flipc(on_state, ctrl_qubits)
    ctrl_push(ctrl_qubits)

    ret = []

    if hasattr(func, '__iter__'):
        for gate in func:
            ret.append(gate(*args, **kwargs))
    else:
        ret = func(*args, **kwargs)

    ctrl_pop()
    if on_state is not None:
        base_flipc(on_state, ctrl_qubits)

    return ret


def _qubit_for_ctrl(qubits: quant | list[quant] | slice | int | list[int]) -> tuple[Callable[[quant], quant] | quant, bool]:  # pylint: disable=line-too-long
    """Get qubits for ctrl"""

    if any(isinstance(qubits, tp) for tp in [slice, int]):
        return lambda q: q[qubits], True
    if hasattr(qubits, '__iter__') and all(isinstance(i, int) for i in qubits):
        return lambda q: q.at(qubits), True
    if hasattr(qubits, '__iter__') and all(isinstance(i, quant) for i in qubits):
        return reduce(add, qubits), False
    return qubits, False


def ctrl(ctrl_qubits: quant | list[quant] | slice | int | list[int],
         func: Callable | list[Callable],
         *args,
         on_state: int | list[int] | None = None,
         later_call: bool = False,
         **kwargs) -> Callable | Any:
    r"""Add control-qubits to a Callable

    :Call with control qubits:

    * ``control`` type must be :class:`~ket.types.quant` or ``[quant]``
    * ``func`` type must be ``Callable`` or ``[Callable]``

    .. code-block:: ket

        ret1 = ctrl(control_qubits, func, *args, **kwargs)
        # Equivalent to:
        # with control(control_qubits):
        #     ret1 = func(*args, **kwargs)

        ret2 = ctrl([q0, q1, q2, ...], func, *args, **kwargs)
        # Equivalent to:
        # with control(*control):
        #     ret2 = func(*args, **kwargs)

        ret3 = ctrl(control_qubits, [f0, f1, f2, ...], *args, **kwargs)
        # Equivalent to:
        # ret3 = []
        # with control(control_qubits):
        #     for f in func:
        #         ret3.append(f(*args, **kwargs))

    :Create controlled-operation:

    1. If the keyword argument ``later_call`` is ``True``, return a
    ``Callable[[], Any]``:

    .. code-block:: ket

        ctrl_func = ctrl(control_qubits, func, *args, **kwargs, later_call=True)
        # Equivalent to:
        # ctrl_func = lambda : ctrl(control_qubits, func, *args, **kwargs)

    Example:

    .. code-block:: ket
        :emphasize-lines: 8

        def increment(q):
            if len(q) > 1:
                ctrl(q[-1], increment, q[:-1])
            X(q[-1])

        size = ceil(log2(len(inputs)))+1
        with quant(size) as inc:
            with around(ctrl(q , increment, inc, later_call=True) for q in inputs):
                with control(inc, on_state=len(inputs)//2):
                    X(output)
            inc.free()

    2. If ``control`` and ``args`` type is ``int``, ``slice``, or
    ``[int]``,  return a ``Callable[[quant], Any]``:

    .. code-block:: ket

        ctrl_func = ctrl(ctrl_index, func, target_index)
        # Equivalent to:
        # ctrl_func = lambda q : ctrl(q[ctrl_index], func, q[target_index])

    Example:

    .. code-block:: ket

        with around(ctrl(0, X, slice(1, None)), q): # ctrl(q[0], X, q[1:])
            H(q[0])                                 # H(q[0])
                                                    # ctrl(q[0], X, q[1:])

    Args:
        control: Control qubits.
        func: Functions to add control.
        args: ``func`` arguments.
        kwargs: ``func`` keyword arguments.
        on_state: Change the control state, same as for :class:`~ket.standard.control`.
        later_call: If ``True``, do not execute and return a ``Callable[[], Any]``.
    """

    ctrl_qubits, create_gate = _qubit_for_ctrl(ctrl_qubits)
    if create_gate:
        target, call_target = _qubit_for_ctrl(*args)

        def _ctrl_gate(qubits: quant) -> quant:
            _ctrl(ctrl_qubits(qubits), func, target(qubits)
                  if call_target else target, on_state=on_state)
            return qubits
        return _ctrl_gate
    if later_call:
        return lambda: _ctrl(ctrl_qubits, func, *args, on_state=on_state, **kwargs)
    return _ctrl(ctrl_qubits, func, *args, on_state=on_state, **kwargs)



#  Copyright 2020, 2021 Evandro Chagas Ribeiro da Rosa <evandro.crr@posgrad.ufsc.br>
#  Copyright 2020, 2021 Rafael de Santiago <r.santiago@ufsc.br>
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from ctypes import *
from .wrapper import load_lib, from_u8_to_str, os_lib_name
from os import environ
from .libket import JSON
import json

DENSE = 0
SPARSE = 1

API_argtypes = {
    'kbw_run_and_set_result': ([c_void_p, c_int32], []),
    'kbw_run_serialized': ([POINTER(c_uint8), c_size_t, POINTER(c_uint8), c_size_t, c_int32, c_int32], [c_void_p]),
    'kbw_result_get': ([c_void_p], [POINTER(c_uint8), c_size_t]),
    'kbw_result_delete': ([c_void_p], []),
}


def kbw_path():
    from os.path import dirname

    if "KBW_PATH" in environ:
        kbw_path = environ["KBW_PATH"]
    else:
        kbw_path = f'{dirname(__file__)}/libs/{os_lib_name("kbw")}'

    return kbw_path


API = load_lib('KBW', kbw_path(), API_argtypes, 'kbw_error_message')

_sim_mode = None


def set_sim_mode_dense():
    global _sim_mode
    _sim_mode = DENSE


def set_sim_mode_sparse():
    global _sim_mode
    _sim_mode = SPARSE


def run_and_set_result(process):
    global _sim_mode

    if _sim_mode is None and 'KBW_MODE' in environ:
        sim_mode = environ['KBW_MODE'].upper()
        if sim_mode == 'DENSE':
            sim_mode = DENSE
        elif sim_mode == 'SPARSE':
            sim_mode = SPARSE
        else:
            raise RuntimeError(
                "undefined value for environment variable 'KBW_MODE', expecting 'DENSE' or 'SPARSE'")
    elif _sim_mode is not None:
        sim_mode = _sim_mode
    else:
        sim_mode = SPARSE

    API['kbw_run_and_set_result'](process, sim_mode)


def run_json(quantum_code, metrics, sim_mode: str = 'sparse'):
    quantum_code = json.dumps(quantum_code).encode()
    quantum_code_size = len(quantum_code)
    metrics = json.dumps(metrics).encode()
    metrics_size = len(metrics)

    sim_mode = {'SPARSE': SPARSE, 'DENSE': DENSE}[sim_mode.upper()]

    result = API['kbw_run_serialized'](
        (c_uint8*quantum_code_size)(*quantum_code),
        quantum_code_size,
        (c_uint8*metrics_size)(*metrics),
        metrics_size,
        JSON,
        sim_mode
    )

    data = json.loads(from_u8_to_str(*API['kbw_result_get'](result)))
    API['kbw_result_delete'](result)

    return data



from __future__ import annotations
#  Copyright 2020, 2021 Evandro Chagas Ribeiro da Rosa <evandro.crr@posgrad.ufsc.br>
#  Copyright 2020, 2021 Rafael de Santiago <r.santiago@ufsc.br>
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from ..base import (base_H, base_phase, base_RX, base_RY, base_RZ,
                    base_S, base_SD, base_T, base_TD, base_X, base_Y, base_Z)
from .quantum_gate import QuantumGate
from ..standard.ctrl import base_flipc
from .base_gates import *

__all__ = ['I', 'X', 'Y', 'Z', 'H', 'S', 'SD', 'T', 'TD', 'phase', 'RX',
           'RY', 'RZ', 'cnot', 'swap', 'RXX', 'RYY', 'RZZ', 'flipc', 'phase_on']

# pylint: disable=line-too-long

I = QuantumGate(
    name='Identity',
    gate=lambda q: q,
    doc={
        'func': '``I(q)``',
        'matrix': r":math:`\begin{bmatrix} 1 & 0 \\ 0 & 1 \end{bmatrix}`",
        'effect': r":math:`\begin{matrix} I\left|0\right> = & \left|0\right> \\ I\left|1\right> = & \left|1\right> \end{matrix}`"
    }
)

X = QuantumGate(
    name='Pauli-X',
    gate=base_X,
    doc={
        'func': '``X(q)``',
        'matrix': r":math:`\begin{bmatrix} 0 & 1 \\ 1 & 0 \end{bmatrix}`",
        'effect': r":math:`\begin{matrix} X\left|0\right> = & \left|1\right> \\ X\left|1\right> = & \left|0\right> \end{matrix}`"
    }
)

Y = QuantumGate(
    name='Pauli-Y',
    gate=base_Y,
    doc={
        'func': '``Y(q)``',
        'matrix': r":math:`\begin{bmatrix} 0 & -i \\ i & 0 \end{bmatrix}`",
        'effect': r":math:`\begin{matrix} Y\left|0\right> = & i\left|1\right> \\ Y\left|1\right> =& -i\left|0\right> \end{matrix}`"
    }
)

Z = QuantumGate(
    name='Pauli-Z',
    gate=base_Z,
    doc={
        'func': '``Z(q)``',
        'matrix': r":math:`\begin{bmatrix}  1 & 0 \\ 0 & -1 \end{bmatrix}`",
        'effect': r":math:`\begin{matrix} Z\left|0\right> = & \left|0\right> \\ Z\left|1\right> = & -\left|1\right> \end{matrix}`"
    }
)

H = QuantumGate(
    name='Hadamard',
    gate=base_H,
    doc={
        'func': '``H(q)``',
        'matrix': r":math:`\frac{1}{\sqrt{2}}\begin{bmatrix} 1 & 1 \\ 1 & -1 \end{bmatrix}`",
        'effect': r":math:`\begin{matrix} H\left|0\right> = & \frac{\left|0\right>+\left|1\right>}{\sqrt{2}} = & \left|+\right> \\ H\left|1\right> = & \frac{\left|0\right>-\left|1\right>}{\sqrt{2}} = & \left|-\right> \\ H\left|+\right> = & \left|0\right> \\ H\left|-\right> = & \left|1\right> \\ \end{matrix}`"
    }
)

S = QuantumGate(
    name='S',
    gate=base_S,
    doc={
        'func': '``S(q)``',
        'matrix': r":math:`\begin{bmatrix} 1 & 0 \\ 0 & i \end{bmatrix}`",
        'effect': r":math:`\begin{matrix}  S\left|0\right> = & \left|0\right> \\  S\left|1\right> = & i\left|1\right> \end{matrix}`"
    }
)

SD = QuantumGate(
    name='S Dagger',
    gate=base_SD,
    doc={
        'func': '``SD(q)``',
        'matrix': r":math:`\begin{bmatrix} 1 & 0 \\ 0 & -i \end{bmatrix}`",
        'effect': r":math:`\begin{matrix} S^\dagger\left|0\right> = & \left|0\right> \\ S^\dagger\left|1\right> = & -i\left|1\right> \end{matrix}`"
    }
)

T = QuantumGate(
    name='T',
    gate=base_T,
    doc={
        'func': '``T(q)``',
        'matrix': r":math:`\begin{bmatrix} 1 & 0 \\ 0 & e^{i\pi/4} \end{bmatrix}`",
        'effect': r":math:`\begin{matrix} T^\dagger\left|0\right> = & \left|0\right> \\ T^\dagger\left|1\right> = & \frac{1+i}{\sqrt{2}}\left|1\right> \end{matrix}`"
    }
)


TD = QuantumGate(
    name='T Dagger',
    gate=base_TD,
    doc={
        'func': '``TD(q)``',
        'matrix': r":math:`\begin{bmatrix} 1 & 0 \\ 0 & e^{-i\pi/4} \end{bmatrix}`",
        'effect': r":math:`\begin{matrix} T^\dagger\left|0\right> = & \left|0\right> \\ T^\dagger\left|1\right> = & \frac{1-i}{\sqrt{2}}\left|1\right> \end{matrix}`"
    }
)

phase = QuantumGate(
    name='Phase',
    gate=base_phase,
    c_args=1,
    doc={
        'func': '``phase(λ, q)``',
        'matrix': r":math:`\begin{bmatrix} 1 & 0 \\ 0 & e^{i\lambda} \end{bmatrix}`",
        'effect': r":math:`\begin{matrix} P\left|0\right> = & \left|0\right> \\ P\left|1\right> =& e^{i\lambda}\left|1\right> \end{matrix}`"
    }
)

RX = QuantumGate(
    name='X-axis Rotation',
    gate=base_RX,
    c_args=1,
    doc={
        'func': '``RX(θ, q)``',
        'matrix': r":math:`\begin{bmatrix} \cos{\frac{\theta}{2}} & -i\sin{\frac{\theta}{2}} \\ -i\sin{\frac{\theta}{2}} & \cos{\frac{\theta}{2}} \end{bmatrix}`",
        'effect': r":math:`\begin{matrix} RX\left|0\right> = & \cos\frac{\theta}{2}\left|0\right> -i\sin\frac{\theta}{2}\left|1\right> \\ RX\left|1\right> =& -i\sin\frac{\theta}{2}\left|0\right> + \cos\frac{\theta}{2}\left|1\right> \end{matrix}`"
    }
)

RY = QuantumGate(
    name='Y-axis Rotation',
    gate=base_RY,
    c_args=1,
    doc={
        'func': '``RY(θ, q)``',
        'matrix': r":math:`\begin{bmatrix} \cos{\frac{\theta}{2}} & -\sin{\frac{\theta}{2}} \\ \sin{\frac{\theta}{2}} & \cos{\frac{\theta}{2}} \end{bmatrix}`",
        'effect': r":math:`\begin{matrix} RY\left|0\right> = & \cos{\theta\over2}\left|0\right> + \sin\frac{\theta}{2}\left|1\right> \\ RY\left|1\right> =& -\sin\frac{\theta}{2}\left|0\right> + \cos\frac{\theta}{2}\left|1\right> \end{matrix}`"
    }
)

RZ = QuantumGate(
    name='Z-axis Rotation',
    gate=base_RZ,
    c_args=1,
    doc={
        'func': '``RZ(θ, q)``',
        'matrix': r":math:`\begin{bmatrix} e^{-i\theta/2} & 0 \\ 0 & e^{i\theta/2} \end{bmatrix}`",
        'effect': r":math:`\begin{matrix} RZ\left|0\right> = & e^{-i\theta/2}\left|0\right> \\ RZ\left|1\right> =& e^{i\theta/2}\left|1\right> \end{matrix}`"
    }
)


cnot = QuantumGate(
    name='Controlled-NOT',
    gate=base_cnot,
    q_args=2,
    doc={
        'func': '``cnot(c, t)``',
        'matrix': r":math:`\begin{bmatrix} 1 & 0 & 0 & 0 \\ 0 & 1 & 0 & 0 \\ 0 & 0 & 0 & 1 \\ 0 & 0 & 1 & 0 \end{bmatrix}`"
    }

)

swap = QuantumGate(
    name='SWAP',
    gate=base_swap,
    q_args=2,
    doc={
        'func': '``swap(a, b)``',
        'matrix': r":math:`\begin{bmatrix} 1 & 0 & 0 & 0 \\ 0 & 0 & 1 & 0 \\ 0 & 1 & 0 & 0 \\ 0 & 0 & 0 & 1 \end{bmatrix}`"
    }

)

RXX = QuantumGate(
    name='XX-axis Rotation',
    gate=base_RXX,
    c_args=1,
    q_args=2,
    doc={
        'func': '``RXX(θ, a, b)``',
        'matrix': r":math:`\begin{bmatrix} \cos\frac{\theta}{2} & 0 & 0 & -i\sin\frac{\theta}{2} \\ 0 & \cos\frac{\theta}{2} & -i\sin\frac{\theta}{2} & 0 \\ 0 & -i\sin\frac{\theta}{2} & \cos\frac{\theta}{2} & 0 \\ -i\sin\frac{\theta}{2} & 0 & 0 & \cos\frac{\theta}{2} \end{bmatrix}`"
    }
)

RYY = QuantumGate(
    name='YY-axis Rotation',
    gate=base_RYY,
    c_args=1,
    q_args=2,
    doc={
        'func': '``RYY(θ, a, b)``',
        'matrix': r":math:`\begin{bmatrix} \cos\frac{\theta}{2} & 0 & 0 & i\sin\frac{\theta}{2} \\ 0 & \cos\frac{\theta}{2} & -i\sin\frac{\theta}{2} & 0 \\ 0 & -i\sin\frac{\theta}{2} & \cos\frac{\theta}{2} & 0 \\ i\sin\frac{\theta}{2} & 0 & 0 & \cos\frac{\theta}{2} \end{bmatrix}`"
    }
)

RZZ = QuantumGate(
    name='ZZ-axis Rotation',
    gate=base_RZZ,
    c_args=1,
    q_args=2,
    doc={
        'func': '``RZZ(θ, a, b)``',
        'matrix': r":math:`\begin{bmatrix} e^{-i \frac{\theta}{2}} & 0 & 0 & 0 \\ 0 & e^{i \frac{\theta}{2}} & 0 & 0 \\ 0 & 0 & e^{i \frac{\theta}{2}} & 0 \\ 0 & 0 & 0 & e^{-i \frac{\theta}{2}} \end{bmatrix}`"
    }
)

flipc = QuantumGate(
    name='Flip to Control State',
    gate=base_flipc,
    c_args=1,
    q_args=None,
    doc={
        'func': '``flipc(state, q)``',
        'matrix': r":math:`\bigotimes_{\text{bit }\in\text{ state}}\begin{cases}X,&\text{for bit }=0\\I,&\text{for bit }=1\end{cases}`"
    }
)

phase_on = QuantumGate(
    name='Phase on State',
    gate=base_phase_on,
    c_args=1,
    q_args=None,
    doc={
        'func': '``phase_on(state, q)``',
        'matrix': r":math:`\sum\delta_k\left|k\right>\left<k\right|,\text{where }\delta_k=\begin{cases}-1,&\text{for }k= \text{state}\\1,&\text{for }k\neq\text{state}\end{cases}`"
    }
)



from setuptools import setup, find_namespace_packages
import os
import shutil


def libs():
    if os.name == 'nt':
        return [('libket', 'ket.dll'), ('kbw', 'kbw.dll')]
    if os.uname().sysname == 'Linux':
        return [('libket', 'libket.so'), ('kbw', 'libkbw.so')]
    if os.uname().sysname == 'Darwin':
        return [('libket', 'libket.dylib'), ('kbw', 'libkbw.dylib')]
    raise OSError('unsupported operational system')


def make_libs():
    try:
        os.mkdir("libs")
    except FileExistsError:
        pass
    os.chdir("libs")
    os.system(
        "git clone https://gitlab.com/quantum-ket/libket.git || git -C libket pull")
    os.system("git clone https://gitlab.com/quantum-ket/kbw.git || git -C kbw pull")

    os.chdir("libket")
    os.system("cargo build --release")
    os.chdir("../kbw")
    os.system("cargo build --release")

    os.chdir("../..")
    try:
        os.mkdir("ket/clib/libs")
    except FileExistsError:
        pass

    print(os.getcwd())
    for dir, lib in libs():
        src = f"libs/{dir}/target/release/{lib}"
        dst = "ket/clib/libs/"+lib
        shutil.copy(src, dst)


make_libs()


def long_description():
    with open('README.md', 'r') as file:
        return file.read()


def version():
    from ket.__version__ import __version__
    return __version__


setup(
    name='ket-lang',
    description='Ket Quantum Programming Language interpreter and library.',
    url='https://quantumket.org',
    version=version(),
    author='Evandro Chagas Ribeiro da Rosa',
    author_email='ev.crr97@gmail.com',
    license='Apache-2.0',
    long_description=long_description(),
    long_description_content_type='text/markdown',
    zip_safe=False,
    include_package_data=True,
    packages=find_namespace_packages(include=['ket', 'ket.*']),
    python_requires='>=3.7',
    classifiers=[
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3 :: Only',
    ],
    entry_points={'console_scripts': ['ket = ket.__main__:__ket__']},
    setup_requires=['wheel'],
)



#  Copyright 2020, 2021 Evandro Chagas Ribeiro da Rosa <evandro.crr@posgrad.ufsc.br>
#  Copyright 2020, 2021 Rafael de Santiago <r.santiago@ufsc.br>
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

# pylint: disable=unused-import, wildcard-import

from .import_ket import _import_globals_ket
from . import *
from .base import label, branch, jump
from .__version__ import __version__
from .preprocessor import *


def __ket__():
    import argparse  # pylint: disable=import-outside-toplevel
    import os  # pylint: disable=import-outside-toplevel

    parser = argparse.ArgumentParser(prog='ket', description='Ket interpreter')
    parser.add_argument('--version', action='version',
                        version=f'Ket {__version__}')

    parser.add_argument('input', metavar='.ket',
                        help='source code', type=str)

    args = parser.parse_args()

    globals()['__name__'] = '__main__'
    globals()['__in_ket__'] = True
    source = os.path.join(os.getcwd(), args.input)
    _import_globals_ket(source, globals())


if __name__ == '__main__':
    __ket__()



from __future__ import annotations
#  Copyright 2020, 2021 Evandro Chagas Ribeiro da Rosa <evandro.crr@posgrad.ufsc.br>
#  Copyright 2020, 2021 Rafael de Santiago <r.santiago@ufsc.br>
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from math import pi
from .clib.libket import libket_dump, libket_future, libket_label, libket_qubit, process
from .clib.libket import (EQ, NEQ, GT, GEQ, LT, LEQ, ADD, SUB, MUL, DIV,
                          SLL, SRL, AND, OR, XOR, PAULI_X, PAULI_Y,
                          PAULI_Z, HADAMARD, PHASE, RX, RY, RZ)


from .clib.wrapper import from_list_to_c_vector

__all__ = ['quant', 'future', 'dump']

# pylint: disable=global-statement, missing-function-docstring, invalid-name


class quant:
    r"""Create list of qubits

    Allocate ``size`` qubits in the state :math:`\left|0\right>`.

    If ``dirty`` is ``True``, allocate ``size`` qubits in an unknown state.

    warning:
        Using dirty qubits may have side effects due to previous entanglements.

    Qubits allocated using the ``with`` statement must be free at the end of the scope.

    Example:

    .. code-block:: ket

        a = H(quant())
        b = X(quant())
        with quant() as aux:
            with around(H, aux):
                with control(aux):
                    swap(a, b)
            result = measure(aux)
            if result == 1:
                X(aux)
            aux.free()

    :Qubit Indexing:

    Use brackets to index qubits as in a ``list`` and use ``+`` to concatenate
    two :class:`~ket.base.quant`.

    Example:

    .. code-block:: ket

        q = quant(20)
        head, tail = q[0], q[1:]
        init, last = q[:-1], q[-1]
        even = q[::2]
        odd = q[1::2]
        reverse = reversed(q) # invert qubits order

        a, b = quant(2) # |a⟩, |b⟩
        c = a+b         # |c⟩ = |a⟩|b⟩

    Args:
        size: The number of qubits to allocate.
        dirty: If ``True``, allocate ``size`` qubits at an unknown state.
        qubits: Initialize the qubit list without allocating. Intended for internal use.
    """

    def __init__(self,
                 size: int = 1,
                 dirty: bool = False,
                 *,
                 qubits: list[libket_qubit] | None = None):
        if qubits is not None:
            self.qubits = qubits
        else:
            self.qubits = [libket_qubit(process_top().allocate_qubit(dirty))
                           for _ in range(size)]

    def __add__(self, other: quant) -> quant:
        return quant(qubits=self.qubits+other.qubits)

    def at(self, index: list[int]) -> quant:
        r"""Return qubits at ``index``

        Create a new :class:`~ket.base.quant` with the qubit references at the
        position defined by the ``index`` list.

        :Example:

        .. code-block:: ket

            q = quant(20)
            odd = q.at(range(1, len(q), 2)) # = q[1::2]

        Args:
            index: List of indexes.
        """

        return quant(qubits=[self.qubits[i] for i in index])

    def free(self, dirty: bool = False):
        r"""Free the qubits

        All qubits must be at the state :math:`\left|0\right>` before the call,
        otherwise set the ``dirty`` param to ``True``.

        Warning:
            No check is applied to see if the qubits are at state
            :math:`\left|0\right>`.

        Args:
            dirty: Set ``True`` to free dirty qubits.
        """

        for qubit in self.qubits:
            process_top().free_qubit(qubit, dirty)

    def is_free(self) -> bool:
        """Return ``True`` when all qubits are free"""
        return all(not qubit.allocated().value for qubit in self.qubits)

    def __reversed__(self):
        return quant(qubits=list(reversed(self.qubits)))

    def __getitem__(self, key):
        qubits = self.qubits.__getitem__(key)
        return quant(qubits=qubits if isinstance(qubits, list) else [qubits])

    class iter:
        """Qubits iterator"""

        def __init__(self, q):
            self.q = q
            self.idx = -1
            self.size = len(q.qubits)

        def __next__(self):
            self.idx += 1
            if self.idx < self.size:
                return self.q[self.idx]
            raise StopIteration

        def __iter__(self):
            return self

    def __iter__(self):
        return self.iter(self)

    def __enter__(self):
        return self

    def __exit__(self, type, value, tb):  # pylint: disable=redefined-builtin, invalid-name
        if not self.is_free():
            raise RuntimeError('non-free quant at the end of scope')

    def __len__(self):
        return len(self.qubits)

    def __repr__(self):
        return f"<Ket 'quant' {[(q.pid().value, q.index().value) for q in self.qubits]}>"


class future:
    """64-bits integer on the quantum computer

    Store a reference to a 64-bits integer available in the quantum computer.

    The integer value are available to the classical computer only after the
    quantum execution.

    The following binary operations are available between
    :class:`~ket.base.future` variables and ``int``:

        ``==``, ``!=``, ``<``, ``<=``,
        ``>``, ``>=``, ``+``, ``-``, ``*``, ``/``, ``<<``, ``>>``, ``and``,
        ``xor``, and ``or``.

    A new :class:`~ket.base.future` variable is created with a (1) quantum
    :func:`~ket.standard.measure`; (2) binary operation with a
    :class:`~ket.base.future`; or (3) directly initialization with a ``int``.

    .. code-block:: ket

        q = H(quant(2))
        a = measure(q) # 1
        b = a*3        # 2
        c = qc_int(42) # 3



    Writing to the attribute ``value`` of a :class:`~ket.base.future` variable
    passes the information to the quantum computer.
    Reading the attribute ``value`` triggers the quantum execution.

    If the test expression of an ``if-then-else`` or ``while`` is type future,
    Ket passes the statement to the quantum computer.

    :Example:

    .. code-block:: ket

        q = quant(2)
        with quant() as aux:
            # Create variable done on the quantum computer
            done = qc_int(False)
            while done != True:
                H(q)
                ctrl(q, X, aux)
                res = measure(aux)
                if res == 0:
                    # Update variable done on the quantum computer
                    done.value = True
                else:
                    X(q+aux)
            aux.free()
        # Get the measurement from the quantum computer
        # triggering the quantum execution
        result = measure(q).value

    """

    def __init__(self, value: int | libket_future):
        if isinstance(value, int):
            self.base = qc_int(value).base
        else:
            self.base = value
        self._value = None

    def __getattr__(self, name):
        if name == "value":
            if self._value is None:
                if not self.available:
                    exec_quantum()
                self._value = self.base.value().value
            return self._value
        return super().__getattribute__(name)

    def __setattr__(self, name, value):
        if name == "value":
            if not isinstance(value, future):
                value = qc_int(value)
            process_top().int_set(self.base, value)
        else:
            super().__setattr__(name, value)

    @property
    def available(self) -> bool:
        return self.base.available().value

    @property
    def index(self) -> int:
        return self.base.index().value

    @property
    def pid(self) -> int:
        return self.base.pid().value

    def __add__(self, other: future | int) -> future:
        if not isinstance(other, future):
            other = qc_int(other)
        return future(libket_future(process_top().add_int_op(ADD, self.base, other.base)))

    def __sub__(self, other: future | int) -> future:
        if not isinstance(other, future):
            other = qc_int(other)
        return future(libket_future(process_top().add_int_op(SUB, self.base, other.base)))

    def __mul__(self, other: future | int) -> future:
        if not isinstance(other, future):
            other = qc_int(other)
        return future(libket_future(process_top().add_int_op(MUL, self.base, other.base)))

    def __truediv__(self, other: future | int) -> future:
        if not isinstance(other, future):
            other = qc_int(other)
        return future(libket_future(process_top().add_int_op(DIV, self.base, other.base)))

    def __floordiv__(self, other: future | int) -> future:
        return self.__truediv__(other)

    def __lshift__(self, other: future | int) -> future:
        if not isinstance(other, future):
            other = qc_int(other)
        return future(libket_future(process_top().add_int_op(SLL, self.base, other.base)))

    def __rshift__(self, other: future | int) -> future:
        if not isinstance(other, future):
            other = qc_int(other)
        return future(libket_future(process_top().add_int_op(SRL, self.base, other.base)))

    def __and__(self, other: future | int) -> future:
        if not isinstance(other, future):
            other = qc_int(other)
        return future(libket_future(process_top().add_int_op(AND, self.base, other.base)))

    def __xor__(self, other: future | int) -> future:
        if not isinstance(other, future):
            other = qc_int(other)
        return future(libket_future(process_top().add_int_op(XOR, self.base, other.base)))

    def __or__(self, other: future | int) -> future:
        if not isinstance(other, future):
            other = qc_int(other)
        return future(libket_future(process_top().add_int_op(OR, self.base, other.base)))

    def __radd__(self, other: future | int) -> future:
        other = qc_int(other)
        return future(libket_future(process_top().add_int_op(ADD, other.base, self.base)))

    def __rsub__(self, other: future | int) -> future:
        other = qc_int(other)
        return future(libket_future(process_top().add_int_op(SUB, other.base, self.base)))

    def __rmul__(self, other: future | int) -> future:
        other = qc_int(other)
        return future(libket_future(process_top().add_int_op(MUL, other.base, self.base)))

    def __rtruediv__(self, other: future | int) -> future:
        other = qc_int(other)
        return future(libket_future(process_top().add_int_op(DIV, other.base, self.base)))

    def __rfloordiv__(self, other: future | int) -> future:
        return self.__rtruediv__(other)

    def __rlshift__(self, other: future | int) -> future:
        other = qc_int(other)
        return future(libket_future(process_top().add_int_op(SLL, other.base, self.base)))

    def __rrshift__(self, other: future | int) -> future:
        other = qc_int(other)
        return future(libket_future(process_top().add_int_op(SRL, other.base, self.base)))

    def __rand__(self, other: future | int) -> future:
        other = qc_int(other)
        return future(libket_future(process_top().add_int_op(AND, other.base, self.base)))

    def __rxor__(self, other: future | int) -> future:
        other = qc_int(other)
        return future(libket_future(process_top().add_int_op(XOR, other.base, self.base)))

    def __ror__(self, other: future | int) -> future:
        other = qc_int(other)
        return future(libket_future(process_top().add_int_op(OR, other.base, self.base)))

    def __lt__(self, other: future | int) -> future:
        if not isinstance(other, future):
            other = qc_int(other)
        return future(libket_future(process_top().add_int_op(LT, self.base, other.base)))

    def __le__(self, other: future | int) -> future:
        if not isinstance(other, future):
            other = qc_int(other)
        return future(libket_future(process_top().add_int_op(LEQ, self.base, other.base)))

    def __eq__(self, other: future | int) -> future:
        if not isinstance(other, future):
            other = qc_int(other)
        return future(libket_future(process_top().add_int_op(EQ, self.base, other.base)))

    def __ne__(self, other: future | int) -> future:
        if not isinstance(other, future):
            other = qc_int(other)
        return future(libket_future(process_top().add_int_op(NEQ, self.base, other.base)))

    def __gt__(self, other: future | int) -> future:
        if not isinstance(other, future):
            other = qc_int(other)
        return future(libket_future(process_top().add_int_op(GT, self.base, other.base)))

    def __ge__(self, other: future | int) -> future:
        if not isinstance(other, future):
            other = qc_int(other)
        return future(libket_future(process_top().add_int_op(GEQ, self.base, other.base)))

    def __repr__(self):
        return f"<Ket 'future' {self.pid, self.index}>"


class dump:
    """Create a snapshot with the current quantum state of ``qubits``

    Gathering any information from a :class:`~ket.base.dump` triggers the quantum execution.

    :Example:

    .. code-block:: ket

        a, b = quant(2)
        with around(cnot(H, I), a, b):
            Y(a)
            inside = dump(a+b)
        outside = dump(a+b)

        print('inside:')
        print(inside.show())
        # inside:
        # |01⟩    (50.00%)
        #          -0.707107i     ≅     -i/√2
        # |10⟩    (50.00%)
        #           0.707107i     ≅      i/√2
        print('outside:')
        print(outside.show())
        # outside:
        # |11⟩    (100.00%)
        #          -1.000000i     ≅     -i/√1

    :param qubits: Qubits to dump.
    """

    def __init__(self, qubits: quant):
        self.base = libket_dump(process_top().dump(
            *from_list_to_c_vector(qubits.qubits)))
        self.qubits = qubits
        self.size = len(qubits)
        self._state = None

    def get_quantum_state(self) -> dict[int, complex]:
        """Get the quantum state

        This function returns a ``dict`` that maps base states to probability amplitude.

        Note:
            Don't use this function if your goal is just to iterate over the basis states.
            Use the attributes :attr:`~ket.base.states`, :attr:`~ket.base.amplitudes`,
            and :attr:`~ket.base.probability` instead.

        :Example:

        .. code-block:: ket

            q = quant(2)
            cnot(H(q[0]), q[1])
            print(dump(q).get_quantum_state())
            # {3: (0.7071067811865476+0j), 0: (0.7071067811865476+0j)}

        """

        if self._state is None:
            self._state = dict(zip(self.states, self.amplitudes))
        return self._state

    @property
    def states(self) -> list[int]:
        """List of basis states"""

        if not self.base.available().value:
            exec_quantum()

        size = self.base.states_size().value

        for i in range(size):
            state, state_size = self.base.state(i)
            yield int(''.join(f'{state[j]:064b}' for j in range(state_size.value)), 2)

    @property
    def amplitudes(self) -> list[complex]:
        """List of probability amplitudes"""

        if not self.available:
            exec_quantum()

        real, size = self.base.amplitudes_real()
        imag, _size = self.base.amplitudes_imag()
        assert size.value == _size.value

        for i in range(size.value):
            yield real[i]+imag[i]*1j

    @property
    def probability(self) -> list[float]:
        """List of measurement probability"""

        for amp in self.amplitudes:
            yield abs(amp)**2

    def show(self, format_str: str | None = None) -> str:
        r"""Return the quantum state as a string

        Use the format starting to change the print format of the basis states:

        * ``i``: print the state in the decimal base
        * ``b``: print the state in the binary base (default)
        * ``i|b<n>``: separate the ``n`` first qubits, the remaining print in the binary base
        * ``i|b<n1>:i|b<n2>[:i|b<n3>...]``: separate the ``n1, n2, n3, ...`` first qubits

        :Example:

        .. code-block:: ket

            q = quant(19)
            X(ctrl(H(q[0]), X, q[1:])[1::2])
            d = dump(q)

            print(d.show('i'))
            # |87381⟩ (50.00%)
            #  0.707107               ≅      1/√2
            # |436906⟩        (50.00%)
            #  0.707107               ≅      1/√2
            print(d.show('b'))
            # |0010101010101010101⟩   (50.00%)
            #  0.707107               ≅      1/√2
            # |1101010101010101010⟩   (50.00%)
            #  0.707107               ≅      1/√2
            print(d.show('i4'))
            # |2⟩|101010101010101⟩    (50.00%)
            #  0.707107               ≅      1/√2
            # |13⟩|010101010101010⟩   (50.00%)
            #  0.707107               ≅      1/√2
            print(d.show('b5:i4'))
            # |00101⟩|5⟩|0101010101⟩  (50.00%)
            #  0.707107               ≅      1/√2
            # |11010⟩|10⟩|1010101010⟩ (50.00%)
            #  0.707107               ≅      1/√2

        Args:
            format: Format string that matches ``(i|b)\d*(:(i|b)\d+)*``.
        """

        if format_str is not None:
            if format_str in ('b', 'i'):
                format_str += str(self.size)
            fmt = []
            count = 0
            for b, size in map(lambda f: (f[0], int(f[1:])), format_str.split(':')):
                fmt.append((b, count, count+size))
                count += size
            if count < self.size:
                fmt.append(('b', count, self.size))
        else:
            fmt = [('b', 0, self.size)]

        def fmt_ket(state, begin, end, f):
            return f'|{state[begin:end]}⟩' if f == 'b' else f'|{int(state[begin:end], base=2)}⟩'

        def state_amp_str(state, amp):
            dump_str = ''.join(
                fmt_ket(f'{state:0{self.size}b}', b, e, f) for f, b, e in fmt)
            dump_str += f"\t({100*abs(amp)**2:.2f}%)\n"
            real = abs(amp.real) > 1e-10
            real_l0 = amp.real < 0

            imag = abs(amp.imag) > 1e-10
            imag_l0 = amp.imag < 0

            sqrt_dem = 1/abs(amp)**2
            use_sqrt = abs(round(sqrt_dem)-sqrt_dem) < .001
            sqrt_dem = f'/√{round(1/abs(amp)**2)}'

            if real and imag:
                sqrt_dem = f'/√{round(2*(1/abs(amp)**2))}'
                sqrt_num = ('(-1' if real_l0 else ' (1') + \
                    ('-i' if imag_l0 else '+i')
                sqrt_str = f'\t≅ {sqrt_num}){sqrt_dem}' if use_sqrt and (
                    abs(amp.real)-abs(amp.real) < 1e-10) else ''
                dump_str += f"{amp.real:9.6f}{amp.imag:+.6f}i"+sqrt_str
            elif real:
                sqrt_num = '  -1' if real_l0 else '   1'
                sqrt_str = f'\t≅   {sqrt_num}{sqrt_dem}' if use_sqrt else ''
                dump_str += f"{amp.real:9.6f}       "+sqrt_str
            else:
                sqrt_num = '  -i' if imag_l0 else '   i'
                sqrt_str = f'\t≅   {sqrt_num}{sqrt_dem}' if use_sqrt else ''
                dump_str += f" {amp.imag:17.6f}i"+sqrt_str

            return dump_str

        return '\n'.join(state_amp_str(state, amp) for state, amp in
                         sorted(zip(self.states, self.amplitudes), key=lambda k: k[0]))

    @property
    def expected_values(self) -> tuple[float, float, float]:
        """X, Y, and Z expected values for one qubit"""

        if self.size != 1:
            raise RuntimeError(
                'Cannot calculate X, Y, and Z expected values from a dump with more than 1 qubit')

        def exp_x(alpha, beta):
            return (beta.conjugate()*alpha+alpha.conjugate()*beta).real

        def exp_y(alpha, beta):
            return (1j*beta.conjugate() * alpha-1j*alpha.conjugate()*beta).real

        def exp_z(alpha, beta):
            return pow(abs(alpha), 2)-pow(abs(beta), 2)

        alpha = 0
        beta = 0
        for a, s in zip(self.amplitudes, self.states):
            if s == 0:
                alpha = a
            else:
                beta = a
        return exp_x(alpha, beta), exp_y(alpha, beta), exp_z(alpha, beta)

    def sphere(self):
        """Result a Bloch sphere

        QuTiP and Matplotlib are required for generating and plotting the sphere.
        """
        try:
            import qutip  # pylint: disable=import-outside-toplevel
        except ImportError as e:
            from sys import stderr  # pylint: disable=import-outside-toplevel
            print("Unable to import QuTiP, try installing:", file=stderr)
            print("\tpip install qutip", file=stderr)
            raise e

        b = qutip.Bloch()
        b.add_vectors(self.expected_values)
        return b

    @property
    def available(self) -> bool:
        return self.base.available().value

    def __len__(self) -> int:
        return self.size

    def __repr__(self) -> str:
        return f"<Ket 'dump' ({repr(self.qubits)})>"


class label:
    """Reference to a code block in the quantum code"""

    def __init__(self):
        self.base = libket_label(process_top().get_label())

    @property
    def index(self) -> int:
        """Return label index"""
        return self.base.index().value

    @property
    def pid(self) -> int:
        """Return process PID"""
        return self.base.pid().value

    def begin(self):
        """Open the code block"""

        process_top().open_block(self.base)

    def __repr__(self) -> str:
        return f"<Ket 'label' {self.pid, self.index}>"


PROCESS_COUNT = 1
PROCESS_STACK = [process(0)]
PROCESS_LAST = None


def process_begin():
    global PROCESS_COUNT
    PROCESS_STACK.append(process(PROCESS_COUNT))
    PROCESS_COUNT += 1


def process_end() -> process:
    global PROCESS_LAST
    PROCESS_LAST = PROCESS_STACK.pop()
    return PROCESS_LAST


def process_top() -> process:
    return PROCESS_STACK[-1]


def process_last() -> process:
    return PROCESS_LAST


QUANTUM_EXECUTION_TARGET = None


def set_quantum_execution_target(func):
    global QUANTUM_EXECUTION_TARGET
    QUANTUM_EXECUTION_TARGET = func


def exec_quantum():
    """Call the quantum execution"""

    process_top().prepare_for_execution()

    error = None
    try:
        QUANTUM_EXECUTION_TARGET(process_end())
    except Exception as e:  # pylint: disable=broad-except
        error = e
    process_begin()
    if error:
        raise error


def qc_int(value: int) -> future:
    """Instantiate an integer on the quantum computer

    args:
        value: Initial value.
    """

    return future(libket_future(process_top().int_new(value)))


def base_measure(qubits: quant) -> future:
    if len(qubits) == 0:
        return None

    size = len(qubits)
    if size <= 64:
        return future(libket_future(process_top().measure(*from_list_to_c_vector(qubits.qubits))))
    return [
        future(libket_future(
            process_top().measure(
                *from_list_to_c_vector(qubits.qubits[i:min(i+63, size)]))
        )) for i in reversed(range(0, size, 63))
    ]


def plugin(name: str, args: str, qubits: quant):
    """Apply plugin

    .. note::

        Plugin availability depends on the quantum execution target.

    args:
        name: Plugin name.
        args: Plugin argument string.
        qubits: Affected qubits.
    """
    process_top().apply_plugin(name.encode(), args.encode(),
                               *from_list_to_c_vector(qubits.qubits))


def ctrl_push(qubits: quant):
    return process_top().ctrl_push(*from_list_to_c_vector(qubits.qubits))


def ctrl_pop():
    return process_top().ctrl_pop()


def adj_begin():
    return process_top().adj_begin()


def adj_end():
    return process_top().adj_end()


def jump(goto: label):
    return process_top().jump(goto.base)


def branch(test: future, then: label, otherwise: label):
    return process_top().branch(test.base, then.base, otherwise.base)


def base_X(qubits: quant):
    for qubit in qubits.qubits:
        process_top().apply_gate(PAULI_X, 0.0, qubit)


def base_Y(qubits: quant):
    for qubit in qubits.qubits:
        process_top().apply_gate(PAULI_Y, 0.0, qubit)


def base_Z(qubits: quant):
    for qubit in qubits.qubits:
        process_top().apply_gate(PAULI_Z, 0.0, qubit)


def base_H(qubits: quant):
    for qubit in qubits.qubits:
        process_top().apply_gate(HADAMARD, 0.0, qubit)


def base_S(qubits: quant):
    for qubit in qubits.qubits:
        process_top().apply_gate(PHASE, pi/2, qubit)


def base_SD(qubits: quant):
    for qubit in qubits.qubits:
        process_top().apply_gate(PHASE, -pi/2, qubit)


def base_T(qubits: quant):
    for qubit in qubits.qubits:
        process_top().apply_gate(PHASE, pi/4, qubit)


def base_TD(qubits: quant):
    for qubit in qubits.qubits:
        process_top().apply_gate(PHASE, -pi/4, qubit)


def base_phase(lambda_, qubits: quant):
    for qubit in qubits.qubits:
        process_top().apply_gate(PHASE, lambda_, qubit)


def base_RX(theta, qubits: quant):
    for qubit in qubits.qubits:
        process_top().apply_gate(RX, theta, qubit)


def base_RY(theta, qubits: quant):
    for qubit in qubits.qubits:
        process_top().apply_gate(RY, theta, qubit)


def base_RZ(theta, qubits: quant):
    for qubit in qubits.qubits:
        process_top().apply_gate(RZ, theta, qubit)
